import datetime
import os
import platform
import re
import socket
import subprocess
import sys

import requests
import serial

if platform.system() == 'Windows':
    pass

from datetime import timedelta
from lxml import etree
from generic_module import Module
from consts import SIWIM_E_VERSION, global_vars, SITE_NAME

ROUTER, IMEI, FREQ_BAND, IMSI, LAT, LON, RSRQ = 'router', 'gsm_imei', 'gsm_frequency_band', 'gsm_imsi', 'gps_lat', 'gps_lon', 'gsm_signal_quality'

class StatusModule(Module):
    def __init__(self, args):
        expected_params = {"time_interval": 600, "host": None, "update_channel": None, "version": None, "url": "/siwim-s/sys_report.php", "protocol": "http", "ntp_conf_path": "", "ovpn_path": "/usr/sbin/openvpn", "save_curr_to_disk": 0,
                           "vpn_host": "172.16.0.1", "root_path": None}
        Module.__init__(self, args, expected_params)
        if platform.system() == 'Windows' and platform.release() == '10':  # This only applies to MK4 systems.
            try:
                self.mk4_com_serial = serial.Serial('COM1')
                self.mk4_com_serial.dtr = False
                self.mk4_com_serial.rts = False
            except serial.serialutil.SerialException as e:
                self.logger.debug(e)

    def VPNstatus(self):
        if platform.system() == "Windows":
            ping = subprocess.Popen(["ping", "-n", "1", "-w", "2", self.vpn_host], stdout=subprocess.PIPE).communicate()[0]
        else:
            ping = subprocess.Popen(["ping", "-c", "1", "-w", "2", self.vpn_host], stdout=subprocess.PIPE).communicate()[0]
        if ('unreachable' not in str(ping)) and ('timed' not in str(ping)) and ('failure' not in str(ping)):
            return "Connected"
        return "Not connected"

    def w32ovpn_version(self):
        sp = subprocess.Popen([self.ovpn_path, "--version"], stdout=subprocess.PIPE)
        output = sp.communicate()[0]
        return output.split(b" ")[1].decode()  # Output of the application is binary and thus has to be split by "binary spaces". We could also decode before splitting, but that's probably redundant.

    def getTeltonika_fw07(self, ip, data):
        login = ip + '/api/login'
        url = ip + '/ubus'

        # A session is required to obtain data, since data is served based on "ubus_rpc_session".
        with requests.Session() as session:
            # Log in to the router and obtain session ID.
            credentials = {
                'username': 'admin',
                'password': 'Siwim323'
            }
            res = session.post(login, json=credentials, timeout=2)
            if res.ok:
                session_id = res.json()['ubus_rpc_session']
            else:
                self.logger.warning('{}: Failed to get session id from router: {}.'.format(res.status_code, res.text))
                return data

            # Obtain device name.
            payload = {
                'jsonrpc': '2.0',
                'method': 'call',
                'params': [session_id, 'uci', 'get', {'config': 'system'}]
            }
            res = session.post(url, json=payload, timeout=2)
            if res.ok:
                result = res.json().get('result')
                data[ROUTER] = result[1]['values']['system']['routername']
            else:
                self.logger.warning('{}: Failed to obtain router name: {}.'.format(res.status_code, res.text))
                return data

            # Obtain mobile data.
            payload = {
                'jsonrpc': '2.0',
                'method': 'call',
                'params': [session_id, 'vuci.network.mobile', 'mobile_info', {'modem': '1-1.4'}]
            }
            res = session.post(url, json=payload, timeout=2)
            if res.ok:
                result = res.json().get('result')[1]['mobile']
                data[IMSI] = result['imsi'].strip()
                data[IMEI] = result['imei'].strip()
                data[RSRQ] = result['rsrq'].strip()
                data[FREQ_BAND] = result['band'].strip()
            else:
                self.logger.warning('{}: Failed to obtain mobile info for {}: {}.'.format(res.status_code, data[ROUTER], res.text))
                return data

            # Obtain GPS coordinates.
            payload = {
                'jsonrpc': '2.0',
                'method': 'call',
                'params': [session_id, 'gpsd', 'position', {}]
            }
            r = session.post(url, json=payload, timeout=2)
            if r.ok:
                result = r.json().get('result')[1]
                data[LAT] = result['latitude']
                data[LON] = result['longitude']
            else:
                self.logger.warning('{}: Failed to obtain GPS info for {}: {}.'.format(res.status_code, data[ROUTER], res.text))
            return data

    def getRouterInfo(self):
        # Define "constants" for convenience
        ports = ('80', '9191')  # Ports that we want to check
        router_id = {'Sierra': 'Sierra', 'Teltonika': '/cgi-bin/luci', 'Teltonika_fw07': 'Teltonika'}  # Teltonika has a redirect, which holds the actual data
        data = {IMSI: 'No IMSI'}
        for port in ports:
            ip = 'http://192.168.4.100:{}'.format(port)  # IP as a variable so that code can be tested anywhere without modifications.
            try:
                req = requests.get(ip, timeout=10)
            except requests.exceptions.ConnectionError:  # If connection can't be established, continue with next port
                continue
            if not req.ok:
                self.logger.info('Router not found at {}: Error code {}. Trying next IP in {}'.format(ip, req.status_code, ', '.join(ports)))
                continue

            router = ''
            for type in router_id.values():
                if req.text.find(type) != -1:  # If the router's name exists in the entry page, we assume it's the router of that company
                    router = type
                    break

            if router == router_id['Sierra']:
                try:
                    data[ROUTER] = '{}: {}'.format(router, re.findall(r'ALEOS Version (?:\d+\.?)+', req.text)[0])  # Exception should happen here if it does.
                    # Obtain the data we're interested in
                    with requests.Session() as s:
                        login_data = '<request xmlns="urn:acemanager"><connect><login>user</login><password><![CDATA[12345]]></password></connect></request>'
                        req = s.post('{}/xml/Connect.xml'.format(ip), data=login_data, timeout=10)
                        if not req.ok:
                            self.logger.warning('Failed to log in to Sierra: {}'.format(req))
                            return
                        # 7=Device model, 10=IMEI, 671=active frequency band, 785=IMSI, 902=latitude, 903=longitude, 10209=signal quality
                        req = s.post('{}/cgi-bin/Embedded_Ace_Get_Task.cgi'.format(ip), data='7,10,671,785,902,903,10209', timeout=10)
                        if not req.ok:
                            self.logger.warning('Failed to obtain information from Sierra: {}'.format(req))
                            return
                    for val, key in zip(req.text.split('!')[:-1], (ROUTER, IMEI, FREQ_BAND, IMSI, LAT, LON, RSRQ)):  # Since it ends with a !, last element is empty
                        data[key] = val.split('=')[1]
                    try:
                        self.logger.info('lat: {}, lon: {}'.format(data[LAT], data[LAT]))
                        data[LAT] = '{}.{}'.format(data[LAT][:-5], data[LAT][-5:])
                        data[LON] = '{}.{}'.format(data[LON][:-5], data[LON][-5:])
                    except:
                        self.logger.warning('Could not convert latitude ({}) and/or longitude ({}) to decimal.'.format(data.get(LAT), data.get(LON)))
                except IndexError:  # The old version of Sierra does not use ALEOS and thus regex match fails
                    self.logger.info('Failed to obtain router information for a Sierra router. Falling back to "old" way.')
                    data[ROUTER] = '{} {}'.format(router, data[ROUTER] if ROUTER in data else '')  # Add the router name before the type.
                    data[LAT], data[LON] = self.gpsLoc()
                    data[IMSI] = self.SierraImsi('12345')

            elif router == router_id['Teltonika']:
                req = requests.get('{}{}'.format(ip, router_id["Teltonika"]), timeout=10)  # This request will return status code 403, but still enough information to identify the router.
                if req.text.find('Teltonika') != -1:  # We double check if it really is Teltonika because their entry page has no unique identification available; we assume it is from redirect path
                    router = 'Teltonika'
                    payload = {
                        'username': 'admin',
                        'password': 'Siwim323'
                    }

                    data[ROUTER] = router
                    data_dict = requests.post('{}/cgi-bin/luci/admin/status/netinfo/mobile?status=1'.format(ip), data=payload, timeout=10).json()
                    data[IMEI] = data_dict['gimei']
                    data[FREQ_BAND] = data_dict['bands']
                    data[IMSI] = data_dict['gimsi']
                    data[RSRQ] = data_dict['grsrq']
                    data_dict = requests.post('{}/cgi-bin/luci/admin/services/gps/get_cord'.format(ip), data=payload, timeout=10).json()
                    data[LAT] = data_dict['latitude']
                    data[LON] = data_dict['longitude']
                else:
                    self.logger.warning('Unrecognised router (assumed Teltonika), IMSI and GPS information not obtained.')
            elif router == router_id['Teltonika_fw07']:
                data = self.getTeltonika_fw07(ip, data)
            else:
                self.logger.warning('Unrecognised router (could not find a match), IMSI and GPS information not obtained.')
            break  # We've gotten our router information, so break the loop.
        self.logger.debug('Router data: {}'.format(data))
        return data

    def gpsLoc(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(5.0)
        data = 'No GPS data'
        try:
            s.connect(('192.168.4.100', 9494))
            data = s.recv(1024).decode()
            s.close()
        except:
            self.logger.debug('Obtaining legacy SierraGpsLoc failed', exc_info=True)

        p = re.compile('GPGGA,([.\d]*),([.\d]*),([NS]),([.\d]*),([EW])')
        a = p.search(data)

        if a == None:
            return (0, 0)
        else:
            latd = float(a.group(2)[:2])
            lond = float(a.group(4)[:3])
            latm = float(a.group(2)[2:])
            lonm = float(a.group(4)[3:])

            lat = latd + (latm / 60)
            lon = lond + (lonm / 60)

            if a.group(3) != 'N':
                lat = -lat
            if a.group(5) != 'E':
                lon = -lon
            return (lat, lon)

    def _recvfind(self, sock, n, to_sec=5):
        d = ''
        absto = datetime.datetime.now() + timedelta(seconds=to_sec)
        while (-1 == d.find(n)) and (datetime.datetime.now() < absto):
            d += sock.recv(256).decode(errors='replace')
        return re.sub(r'[^a-zA-Z0-9 \n]', '', d)

    def SierraImsi(self, pwd):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(5.0)
        imsi = 'No IMSI'
        try:
            s.connect(('192.168.4.100', 2332))
            data = self._recvfind(s, ':')
            if data.find('login') >= 0:
                # print 'GX'
                s.sendall('user\r\n'.encode())
                data = self._recvfind(s, ':')
                # print 'MP'
            s.sendall((pwd + '\r\n').encode())
            data = self._recvfind(s, 'OK')
            s.sendall('at+cimi?\r\n'.encode())
            data = self._recvfind(s, 'OK')
            data = data.split()
            if len(data) > 1: imsi = data[-2]
            s.close()
        except:
            self.logger.debug('Obtaining legacy SierraImsi failed', exc_info=True)
        return imsi

    def swmiver(self):
        return ".".join(str(x) for x in self.version)

    def ntpsvr(self, loc):
        fn = loc
        if not os.path.exists(fn):
            return 'missing'
        f = open(fn)
        try:
            lines = f.readlines()
        finally:
            f.close()
        for l in lines:
            if l.strip().startswith('server'):
                return l.strip().split()[1]
        for l in lines:
            if l.strip().startswith('pool'):
                return l.strip().split()[1]
        return 'unknown'

    def swmstat(self):
        kvp = {}
        try:
            kvp['python_version'] = str(sys.version_info[0]) + '.' + str(sys.version_info[1]) + '.' + str(sys.version_info[2])
        except:
            self.logger.warning('Python version not saved.')
        try:
            kvp['siwim_version'] = '4' if sys.getwindowsversion()[0] == 10 else '3'
            if sys.getwindowsversion()[0] == 10 and sys.version_info[0] == 2:  # Log error if the system is MK4, but is running on Python2
                self.logger.warning('Version mismatch! Windows 10 should use Python3.')
        except:
            self.logger.warning('SiWIM version unavailable.')
        try:
            kvp['update_channel'] = self.update_channel
        except:
            self.logger.warning('Update channel not saved.')
        try:
            kvp['usb_info'] = subprocess.check_output([sys.executable, 'get_usb_info.py'], stderr=open(os.devnull, 'w')).decode()
        except:
            self.logger.warning('USB info not saved.')
        try:
            kvp['timestamp'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')
        except:
            self.logger.warning('Timestamp not saved.')
        try:
            kvp['computer_name'] = subprocess.check_output(['hostname']).strip().decode().upper()
        except:
            self.logger.warning('Computer name not saved.')
        # To reduce repeated code, map values of partitions depending on platform and save them to partitions dict.
        # i like having declarations outside indented code blocks
        partitions = {}
        if platform.system() == 'Windows':
            partitions = {'disk_c': 'c:', 'disk_d': 'd:', 'disk_e': 'e:'}
        else:
            partitions = {'disk_c': '/', 'disk_d': self.root_path, 'disk_e': '/'}
        for key, p in partitions.items():
            try:
                kvp['{0}_free'.format(key)] = self.get_free_mb(p)
            except:
                self.logger.warning('Device {0} not found.'.format(p))
        try:
            kvp['siwim_site_name'] = global_vars[SITE_NAME]
        except:
            self.logger.warning('Site name not saved.')
        try:
            kvp['vpn_status'] = self.VPNstatus()
            if platform.system() == 'Windows' and platform.release() == '10':
                try:
                    self.mk4_com_serial.rts = True if kvp['vpn_status'] == 'Connected' else False
                except AttributeError:  # If self.mk4_com_serial does not exist, it's already being set in another Status module, so we don't care.
                    pass
        except:
            self.logger.warning('VPN status not saved.')
        try:
            kvp[SIWIM_E_VERSION] = global_vars[SIWIM_E_VERSION]
        except:
            self.logger.warning('E version not saved.')
        try:
            kvp['siwim_i_version'] = self.swmiver()
        except:
            self.logger.warning('I version not saved.')
        for key, p in partitions.items():
            try:
                kvp['{0}_total'.format(key)] = self.get_total_mb(p)
            except:
                self.logger.warning('Total MB not saved for {0}.'.format(p))
        try:
            tmp = self.getRouterInfo()
            for key, val in tmp.items():
                kvp[key] = val
        except:
            self.logger.warning('GSM IMSI and location not saved.')
        try:
            if self.ntp_conf_path == "" or self.ntp_conf_path is None:
                if platform.system() == 'Windows':
                    kvp['ntp_server'] = self.ntpsvr("c:/windows/system32/drivers/etc/ntp.conf")
                else:
                    kvp['ntp_server'] = self.ntpsvr("/etc/ntp.conf")
            else:
                kvp['ntp_server'] = self.ntpsvr(self.ntp_conf_path)
        except:
            self.logger.warning('NTP server not saved')
        try:
            kvp['openvpn_version'] = self.w32ovpn_version()
        except:
            self.logger.warning('OpenVPN not saved.')
        if platform.system() == 'Windows' and platform.release() == '10':
            try:
                # Check if task siwim_mcp.exe is running.
                call = 'TASKLIST', '/FI', 'imagename eq siwim_mcp.exe'
                output = subprocess.check_output(call).decode()
                last_line = output.strip().split('\r\n')[-1]
                self.mk4_com_serial.dtr = last_line.lower().startswith('siwim_mcp.exe')
            except AttributeError:  # If the port does not exist, simply skip it. This occurs if there's more than one status module.
                pass
            except:
                self.logger.warning('Failed to get info on SiWIM-E.')
                self.mk4_com_serial.dtr = False
        # Add all keys that have been sent upstream.
        for key, val in self.upstream_info.items():
            try:
                kvp[key] = val
            except:
                self.logger.warning(str(key) + ' not saved.')
        return kvp

    def run(self):
        self.alive = True
        self.end = False
        try:
            while True:
                if self.end:
                    self.alive = False
                    self.logger.debug('Thread closed correctly.')
                    return
                try:
                    kvp = self.swmstat()
                    root = etree.Element("swmstat")

                    voltage_present = False
                    for key, val in kvp.items():
                        if isinstance(val, dict):
                            nest = root.find(key[1] + "s")
                            if nest == None:
                                nest = etree.Element(key[1] + "s")
                                root.append(nest)
                            node = etree.Element(key[1])
                            nest.append(node)
                            for ky in val.keys():
                                node.attrib[ky] = val[ky]
                        else:
                            # ugly hack, but alas
                            nn = ""
                            if isinstance(key, tuple):
                                nn = key[1]
                            else:
                                nn = key
                            node = etree.Element(nn)
                            node.text = str(val)
                            root.append(node)
                            if nn == "siwim_system_voltage":
                                voltage_present = True
                    # for non-standard systems that don't yield voltage info, despite it being mandatory for a valid register event(alive)
                    # if this piece of info ever stops being mandatory (i see no reason why it is), remove this hack
                    if not voltage_present:
                        node = etree.Element("siwim_system_voltage")
                        node.text = "-1"
                        root.append(node)
                    if self.save_curr_to_disk == 1:
                        self.logger.info('Writing status to disk ...')
                        fd = open("log/swmstat.xml", "w")
                        fd.write(etree.tostring(root).decode())
                        fd.close()
                    resp = requests.post('{0}://{1}{2}'.format(self.protocol, self.host, self.url), {'sys': kvp["computer_name"].lower(), 'msg': etree.tostring(root).decode()}, timeout=10, verify=False)
                    if resp.status_code != 204:  # Log, if response isn't empty.
                        self.logger.warning('Response: {0}'.format(resp.text))
                except (socket.error, requests.ConnectTimeout) as e:
                    self.logger.warning('Connection attempt failed ({0}).'.format(e))
                except:
                    self.logger.exception('General error:')
                self.zzzzz(self.time_interval)
        except:
            self.logger.exception('Fatal error:')
            self.alive = False
            if platform.system() == 'Windows' and platform.release() == '10':
                self.mk4_com_serial.close()
